#include "defines.h"

#ifdef MALLOC_LOG

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

#include "global.h"

struct structMallocLog
{
	void* Pointer;
	char Ident[256];
	int Size;
	int ID;
	LARGE_INTEGER TimeStamp;
	int vAlloc;
};

char OperatorNewIdent[256];

structMallocLog MallocLogs[MAX_MALLOC];
int MallocLogCount = 0;
int MallocID = 0;

extern "C" int MaxAlloc;
int MaxAlloc = 0;

int VirtualCompactCur = VIRTUAL_COMPACT_BASE;

struct structPersistend
{
	int PersistendID;
	int PersistendSize;
};

structPersistend* Persistends;
int PersistendIDCount = 0;
int CurMallocID = 0;

void __cdecl ListAlloc(int Start)
{
	int i;

	bool found = false;

	if (MallocLogCount > Start)
	{
		for (i = Start;i <= MallocLogCount;i++)
		{
			if (found == false)
			{
				printf("Memmory Leak Found\n\n");
				found = true;
			}
			printf("\t%d\t:%30s (ID: %8d - Size: %5d)  ", i, MallocLogs[i].Ident, MallocLogs[i].ID, MallocLogs[i].Size);
			if (MallocLogs[i].Size < 50)
			{
				((char*) MallocLogs[i].Pointer)[MallocLogs[i].Size - 1] = 0;
				printf("%s", MallocLogs[i].Pointer);
			}
			printf("\n");
		}
		if (found)
		{
			breakpoint("ListAlloc");
		}
	}
}

void* __cdecl mallocloga(int size, char* Ident)
{
	void *tmp, *Pointer;
	int FoundPage = 0;
	int ContinousBlocks = 0;
	int ReserveHigh2 = 0;

	if (DebugOutput || AllocOutput || DEBUG_ALLOC)
	{
		printf("Allocating %d Bytes (%s)\n", size, Ident);
	}

	if (ReserveHigh)
	{
		if (Persistends[CurMallocID].PersistendID == MallocID)
		{
			CurMallocID++;
			ReserveHigh2 = 1;
		}
	}

	if (ReserveHigh2)
	{
		Pointer = tmp = (void*) VirtualCompactCur;
		VirtualCompactCur += (size = Persistends[CurMallocID - 1].PersistendSize);
	}
	else
	{
		tmp = malloc(size);
		Pointer = tmp;
	}

	if (tmp == 0)
	{
		printf("Memmory Allocation Error\n");
		breakpoint("mallocloga");
	}

	if (MallocLogCount < MAX_MALLOC)
	{
		MallocLogCount++;
		if (MallocLogCount > MaxAlloc)
		{
			MaxAlloc = MallocLogCount;
		}
		MallocLogs[MallocLogCount].ID = MallocID++;
		MallocLogs[MallocLogCount].Pointer = Pointer;
		MallocLogs[MallocLogCount].Size = size;
		MallocLogs[MallocLogCount].vAlloc = ReserveHigh2;
		QueryPerformanceCounter(&MallocLogs[MallocLogCount].TimeStamp);
		strcpy(MallocLogs[MallocLogCount].Ident, Ident);
	}
	else
	{
		printf("Could not log Memory Allocation\n");
		breakpoint("mallocloga");
	}
	return(Pointer);
}

void __cdecl freeloga(void* Pointer)
{
	int i;
	bool found = false;
	int vAllocated;
	int nAllocated;

	for (i = MallocLogCount;i > 0;i--)
	{
		if (MallocLogs[i].Pointer == Pointer)
		{
			if (DebugOutput || AllocOutput || DebugOutput)
			{
				printf("Freeing %s\n", MallocLogs[i].Ident);
			}
			vAllocated = MallocLogs[i].vAlloc;
			nAllocated = MallocLogs[i].Size;
			if (i < MallocLogCount)
			{
				CopyMemory(&MallocLogs[i], &MallocLogs[MallocLogCount], sizeof(structMallocLog));
			}
			found = true;
			MallocLogCount--;
			break;
		}
	}
	if (!found)
	{
		printf("Freed On non allocated Block\n");
		breakpoint("freeloga");
	}
	else
	{
		if (!vAllocated)
		{
			free(Pointer);
		}
	}
}

void* __cdecl reallocloga(void* Pointer, int Size)
{
	int i;
	for (i = MallocLogCount;i > 0;i--)
	{
		if (MallocLogs[i].Pointer == Pointer)
		{
			if (MallocLogs[i].vAlloc)
			{
				if (Size > MallocLogs[i].Size)
				{
					breakpoint("Cannot Realloc Dump Memory");
				}
				else
				{
					return(Pointer);
				}
			}
			Pointer = realloc(MallocLogs[i].Pointer, Size);
			if (Pointer == 0)
			{
				breakpoint("Memory Allocation Error");
			}
			if (Size > MallocLogs[i].Size)
			{
				MallocLogs[i].Size = Size;
			}
			MallocLogs[i].Pointer = Pointer;
			if (strlen(MallocLogs[i].Ident) < 240)
			{
				strcat(MallocLogs[i].Ident, "-re");
			}
			else if (strstr(MallocLogs[i].Ident + 239, ".") == 0)
			{
				strcat(MallocLogs[i].Ident, "...");
			}
			return(Pointer);
		}
	}
	breakpoint("Cannot realloc unallocated Memory");
	return(NULL);
}

int __cdecl SaveAllocs(int CompactSize, __int64 Start, __int64 End, char* FileName, void* base, unsigned int base_size, ...)
{
	HANDLE hFile;
	unsigned long int i, j;

	unsigned int BaseAddr = -1, EndAddr = 0, AddrCount = 0;

	printf("Dumping Source...");

	hFile = CreateFile(FileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		printf("Could not Open File: %s\n", FileName);
		return(1);
	}

	WriteFile(hFile, &CompactSize, sizeof(int), &i, NULL);
	WriteFile(hFile, (void*) VIRTUAL_COMPACT_BASE, CompactSize, &j, NULL);

	if (i != sizeof(int) || j != (unsigned) CompactSize)
	{
		printf("Error Writing File\n");
		CloseHandle(hFile);
		return(1);
	}

	WriteFile(hFile, &base_size, sizeof(int), &i, NULL);
	WriteFile(hFile, base, base_size, &j, NULL);
	if (i != sizeof(int) || j != base_size)
	{
		printf("Error Writing File\n");
		CloseHandle(hFile);
		return(1);
	}

	
	va_list valist;
	unsigned int curarg, curarg2;
	va_start(valist, base_size);
	curarg = va_arg(valist, int);
	while (curarg)
	{
		curarg2 = va_arg(valist, int);
		WriteFile(hFile, (void*) &curarg2, sizeof(int), &i, NULL);
		WriteFile(hFile, (void*) curarg, curarg2, &j, NULL);
		if (i != sizeof(int) || j != curarg2)
		{
			printf("Error Writing File\n");
			CloseHandle(hFile);
			return(1);
		}
		curarg = va_arg(valist, int);
	}
	va_end(valist);
	
	CloseHandle(hFile);

	printf("Done\n\n");

	return(0);
}

int __cdecl q_compare_int(const void* a, const void* b)
{
	return (((structPersistend*) a)->PersistendID - ((structPersistend*) b)->PersistendID);
}

int __cdecl PackMallocIDS(int start)
{
	int i;
	int PersistendSize = 0;

	PersistendIDCount = 0;
	for (i = 1;i <= MallocLogCount;i++)
	{
		if (MallocLogs[i].ID >= start && MallocLogs[i].ID < MallocID)
		{
			PersistendIDCount++;
			PersistendSize += MallocLogs[i].Size;
		}
	}

	Persistends = (structPersistend*) malloc(PersistendIDCount * sizeof(structPersistend));
	PersistendIDCount = 0;

	for (i = 1;i <= MallocLogCount;i++)
	{
		if (MallocLogs[i].ID >= start && MallocLogs[i].ID < MallocID)
		{
			Persistends[PersistendIDCount].PersistendID = MallocLogs[i].ID;
			Persistends[PersistendIDCount].PersistendSize = MallocLogs[i].Size;
			PersistendIDCount++;
		}
	}
	CurMallocID = 0;

	qsort(Persistends, PersistendIDCount, sizeof(structPersistend), &q_compare_int);

	return(PersistendSize);
}

void __cdecl ReleaseMallocIDS()
{
	free(Persistends);
	PersistendIDCount = 0;
	CurMallocID = 0;
}

void* __cdecl LimitedAlloc(unsigned int size, int AttachBlocks)
{
	static int used = 0;
	if (used)
	{
		breakpoint("LimitedAlloc");
	}
	used = 1;

	unsigned int allocsize = size % dwPageSize;
	allocsize = dwPageSize - allocsize;
	allocsize += size;

	unsigned int reservesize = allocsize % 65536;
	reservesize = 65536 - reservesize;
	reservesize += allocsize + AttachBlocks * 65536;

	int rVal;
	void* ptr;
	rVal = (int) VirtualAlloc((void*) (900 * 1024 * 1024), reservesize, MEM_RESERVE, PAGE_NOACCESS);
	rVal = (int) VirtualAlloc((void*) (900 * 1024 * 1024), allocsize, MEM_COMMIT, PAGE_READWRITE);
	rVal = (int) VirtualAlloc((void*) (900 * 1024 * 1024 + allocsize), reservesize - allocsize, MEM_COMMIT, PAGE_NOACCESS);
	ptr = (void*) (900 * 1024 * 1024 + allocsize - size);

	return(ptr);
}

void* __cdecl operator new[] (size_t size)
{
	void* tmp;
	CopyMemory(OperatorNewIdent, "new[] ", 6);
	tmp = mallocloga(size, OperatorNewIdent);
	OperatorNewIdent[6] = 0;
	return(tmp);
}

void* __cdecl operator new (size_t size)
{
	void* tmp;
	CopyMemory(OperatorNewIdent, "new   ", 6);
	tmp = mallocloga(size, OperatorNewIdent);
	OperatorNewIdent[6] = 0;
	return(tmp);
}

void __cdecl operator delete[] (void* pointer)
{
	freeloga(pointer);
}

void __cdecl operator delete (void* pointer)
{
	freeloga(pointer);
}

#endif